/********************************************************************/
/*																	*/
/*						Portable MK14 emulator in 'C'				*/
/*																	*/
/*								CPU Emulator						*/
/*																	*/
/********************************************************************/

#include "scmp.h"

int Acc,Ext,Stat;						/* SC/MP CPU Registers */
int Ptr[4];
unsigned char Memory[4096];				/* SC/MP Program Memory */
long Cycles;							/* Cycle Count */

static int Indexed(int);                /* Local prototypes */
static int AutoIndexed(int);
static int BinAdd(int,int);
static int DecAdd(int,int);

/********************************************************************/
/*							Reset the CPU							*/
/********************************************************************/

void ResetCPU(void)
{
Acc = Ext = Stat = 0;					/* Zero all registers */
Cycles = 0L;
Ptr[0] = Ptr[1] = Ptr[2] = Ptr[3] = 0;
}

/********************************************************************/
/*					  Execute a block of code						*/
/********************************************************************/

#define CYCLIMIT	(10000L)

void BlockExecute(void)
{
Execute(8192);							/* Do opcodes until cyclimit */

if (CONKeyPressed(KEY_RESET))			/* Check for CPU Reset */
						ResetCPU();
}

/********************************************************************/
/*			Execute a given number of instructions					*/
/********************************************************************/

#define CYC(n)	Cycles+= (long)(n)		/* Bump the cycle counter */

										/* Shorthand for multiple case */
#define CAS4(n) case n: case n+1: case n+2: case n+3
#define CAS3(n) case n: case n+1: case n+2

#define CM(n)	((n) ^ 0xFF)      		/* 1's complement */

void Execute(int Count)
{
register int Opcode;
register int Pointer;
int n;
long l;

while (Count-- > 0)
	{
	while (Cycles > CYCLIMIT)			/* Check for cycle limit */
		{
		Latency();
		Cycles = Cycles - CYCLIMIT;
		CONSynchronise(CYCLIMIT);
		}

	FETCH(Opcode);						/* Fetch the opcode, hack of the */
	Pointer = Opcode & 3;				/* pointer reference */

	switch(Opcode)						/* Pointer instructions first */
		{
										/* LD (Load) */
		CAS4(0xC0):	Acc = ReadMemory(Indexed(Pointer));CYC(18);break;
		case 0xC4:	FETCH(Acc);CYC(10);break;
		CAS3(0xC5):	Acc = ReadMemory(AutoIndexed(Pointer));CYC(18);break;
		case 0x40:	Acc = Ext;CYC(6);break;

										/* ST (Store) */
		CAS4(0xC8):	WriteMemory(Indexed(Pointer),Acc);CYC(18);break;
		CAS3(0xCD):	WriteMemory(AutoIndexed(Pointer),Acc);CYC(18);break;

										/* AND (And) */
		CAS4(0xD0):	Acc = Acc & ReadMemory(Indexed(Pointer));CYC(18);break;
		case 0xD4:	FETCH(n);Acc = Acc & n;CYC(10);break;
		CAS3(0xD5):	Acc = Acc & ReadMemory(AutoIndexed(Pointer));CYC(18);break;
		case 0x50:	Acc = Acc & Ext;CYC(6);break;

										/* OR (Or) */
		CAS4(0xD8):	Acc = Acc | ReadMemory(Indexed(Pointer));CYC(18);break;
		case 0xDC:	FETCH(n);Acc = Acc | n;CYC(10);break;
		CAS3(0xDD):	Acc = Acc | ReadMemory(AutoIndexed(Pointer));CYC(18);break;
		case 0x58:	Acc = Acc | Ext;CYC(6);break;

										/* XOR (Xor) */
		CAS4(0xE0):	Acc = Acc ^ ReadMemory(Indexed(Pointer));CYC(18);break;
		case 0xE4:	FETCH(n);Acc = Acc ^ n;CYC(10);break;
		CAS3(0xE5):	Acc = Acc ^ ReadMemory(AutoIndexed(Pointer));CYC(18);break;
		case 0x60:	Acc = Acc ^ Ext;CYC(6);break;

										/* DAD (Dec Add) */
		CAS4(0xE8):	Acc = DecAdd(Acc,ReadMemory(Indexed(Pointer)));CYC(23);break;
		case 0xEC:	FETCH(n);Acc = DecAdd(Acc,n);CYC(15);break;
		CAS3(0xED):	Acc = DecAdd(Acc,ReadMemory(AutoIndexed(Pointer)));CYC(23);break;
		case 0x68:	Acc = DecAdd(Acc,Ext);CYC(11);break;

										/* ADD (Add) */
		CAS4(0xF0):	Acc = BinAdd(Acc,ReadMemory(Indexed(Pointer)));CYC(19);break;
		case 0xF4:	FETCH(n);CYC(11);Acc = BinAdd(Acc,n);break;
		CAS3(0xF5):	CYC(19);Acc = BinAdd(Acc,ReadMemory(AutoIndexed(Pointer)));break;
		case 0x70:	Acc = BinAdd(Acc,Ext);CYC(7);break;

										/* CAD (Comp Add) */
		CAS4(0xF8):	Acc = BinAdd(Acc,CM(ReadMemory(Indexed(Pointer))));CYC(20);break;
		case 0xFC:	FETCH(n);CYC(12);Acc = BinAdd(Acc,CM(n));break;
		CAS3(0xFD):	Acc = BinAdd(Acc,CM(ReadMemory(AutoIndexed(Pointer))));CYC(20);break;
		case 0x78:	Acc = BinAdd(Acc,CM(Ext));CYC(8);break;

		CAS4(0x30):						/* XPAL */
			n = Ptr[Pointer];CYC(8);
			Ptr[Pointer] = (n & 0xFF00) | Acc;
			Acc = n & 0xFF;
			break;
		CAS4(0x34):						/* XPAH */
			n = Ptr[Pointer];CYC(8);
			Ptr[Pointer] = (n & 0xFF) | (Acc << 8);
			Acc = (n >> 8) & 0xFF;
			break;
		CAS4(0x3C):						/* XPPC */
			n = Ptr[Pointer];Ptr[Pointer] = Ptr[0];Ptr[0] = n;
			CYC(7);break;

		CAS4(0x90):							/* Jumps */
			CYC(11);Ptr[0] = Indexed(Pointer);break;
		CAS4(0x94):
			CYC(11);n = Indexed(Pointer);
			if ((Acc & 0x80) == 0) Ptr[0] = n;
			break;
		CAS4(0x98):
			CYC(11);n = Indexed(Pointer);if (Acc == 0) Ptr[0] = n;
			break;
		CAS4(0x9C):
			CYC(11);n = Indexed(Pointer);if (Acc != 0) Ptr[0] = n;
			break;

		CAS4(0xA8):							/* ILD and DLD */
			n = Indexed(Pointer);Acc = (ReadMemory(n)+1) & 0xFF;
			CYC(22);WriteMemory(n,Acc);break;

		CAS4(0xB8):
			n = Indexed(Pointer);Acc = (ReadMemory(n)-1) & 0xFF;
			CYC(22);WriteMemory(n,Acc);break;

		case 0x8F:							/* DLY */
			FETCH(n);l = ((long)n) & 0xFFL;
			l = 514L * l + 13 + Acc; Acc = 0xFF;
			CYC(l);break;

		case 0x01:							/* XAE */
			n = Acc;Acc = Ext;Ext = n;break;
		case 0x19:							/* SIO */
			CYC(5);Ext = (Ext >> 1) & 0x7F;break;
		case 0x1C:							/* SR */
			CYC(5);Acc = (Acc >> 1) & 0x7F;break;
		case 0x1D:							/* SRL */
			Acc = (Acc >> 1) & 0x7F;
			CYC(5);Acc = Acc | (Stat & 0x80);break;
		case 0x1E:							/* RR */
			n = Acc;Acc = (Acc >> 1) & 0x7F;
			if (n & 0x1) Acc = Acc | 0x80;
			CYC(5);break;
		case 0x1F:							/* RRL */
			n = Acc;Acc = (Acc >> 1) & 0x7F;
			if (Stat & 0x80) Acc = Acc | 0x80;
			Stat = Stat & 0x7F;
			if (n & 0x1) Stat = Stat | 0x80;
			CYC(5);break;

		case 0x00:							/* HALT */
			CYC(8);break;
		case 0x02:							/* CCL */
			Stat &= 0x7F;CYC(5);break;
		case 0x03:							/* SCL */
			Stat |= 0x80;CYC(5);break;
		case 0x04:                    		/* DINT */
			Stat &= 0xF7;CYC(6);break;
		case 0x05:							/* IEN */
			Stat |= 0x08;CYC(6);break;
		case 0x06:							/* CSA */
			Acc = Stat;CYC(5);break;
		case 0x07:                   		/* CAS */
			Stat = Acc & 0xCF;CYC(6);break;
		case 0x08:							/* NOP */
			break;
		}
	}
}

/********************************************************************/
/*							  Decimal Add							*/
/********************************************************************/

static int DecAdd(int v1,int v2)
{
int n1 = (v1 & 0xF) + (v2 & 0xF);			/* Add LSB */
int n2 = (v1 & 0xF0) + (v2 & 0xF0);			/* Add MSB */
if (Stat & 0x80) n1++;						/* Add Carry */
Stat = Stat & 0x7F;							/* Clear CYL */
if (n1 > 0x09)								/* Digit 1 carry ? */
	{
	n1 = n1 - 0x0A;
	n2 = n2 + 0x10;
	}
n1 = (n1 + n2);
if (n1 > 0x99)								/* Digit 2 carry ? */
	{
	n1 = n1 - 0xA0;
	Stat = Stat | 0x80;
	}
return(n1 & 0xFF);
}

/********************************************************************/
/*							  Binary Add							*/
/********************************************************************/

#define SGN(x) ((x) & 0x80)

static int BinAdd(int v1,int v2)
{
int n;
n = v1 + v2 + ((Stat & 0x80) ? 1 : 0);	/* Add v1,v2 and carry */
Stat = Stat & 0x3F;						/* Clear CYL and OV */
if (n & 0x100) Stat = Stat | 0x80;		/* Set CYL if required */
if (SGN(v1) == SGN(v2) &&				/* Set OV if required */
			SGN(v1) != SGN(n)) Stat |= 0x40;
return(n & 0xFF);
}

/********************************************************************/
/*							Indexing Mode							*/
/********************************************************************/

static int Indexed(int p)
{
int Offset;
FETCH(Offset);							/* Get offset */
if (Offset == 0x80) Offset = Ext;		/* Using 'E' register ? */
if (Offset & 0x80) Offset = Offset-256;	/* Sign extend */
return(ADD12(Ptr[p],Offset));			/* Return result */
}

/********************************************************************/
/*						  Auto-indexing mode						*/
/********************************************************************/

static int AutoIndexed(int p)
{
int Offset,Address;
FETCH(Offset);							/* Get offset */
if (Offset == 0x80) Offset = Ext;		/* Using E ? */
if (Offset & 0x80) Offset = Offset-256;	/* Sign extend */
if (Offset < 0)							/* Pre decrement on -ve offset */
	Ptr[p] = ADD12(Ptr[p],Offset);
Address = Ptr[p];						/* The address we're using */
if (Offset > 0)							/* Post increment on +ve offset */
	Ptr[p] = ADD12(Ptr[p],Offset);
return(Address);
}
